"""Configuration format loaders"""
import os
from abc import ABC, abstractmethod
import yaml
from pydantic import create_model


def load_configuration(configuration_file_path, parameters_file_path, bundles):
    """Combines the configuration and parameters and build the configuration object"""
    mappings = {}
    for bundle in bundles:
        if hasattr(bundle, "config_mapping"):
            mappings.update(bundle.config_mapping)
    loader = YmlLoader()
    return loader.build_config(mappings, config_source=configuration_file_path, parameters_source=parameters_file_path)


def is_string(value):
    """Check if the value is actually a string or not"""
    try:
        float(value)
        return False
    except ValueError:
        if value.lower() in ["true", "false"]:
            return False
        return True


class ConfigurationLoader(ABC):
    """Base configuration loader"""
    @abstractmethod
    def load_parameters(self, source):
        """Convert the source into a dictionary"""

    @abstractmethod
    def load_config(self, config_source, parameters_source):
        """Prase the config file and build a dictionary"""

    def build_config(self, config_mappings, config_source, parameters_source):
        """By using the loaded parameters and loaded config, build the final configuration object"""
        configuration_class = create_model('Configuration', **{k: (v, ...) for k, v in config_mappings.items()})
        return configuration_class(**self.load_config(config_source, parameters_source))


class YmlLoader(ConfigurationLoader):
    """YML Format parser and config loader"""
    def load_parameters(self, source):
        """For YML, the source it the file path"""
        with open(source) as parameters_source:
            loaded = yaml.safe_load(parameters_source.read())
            if loaded:
                for key, value in loaded.items():
                    if isinstance(value, str):
                        loaded[key] = "'" + value + "'"
                return loaded
        return {}

    def load_config(self, config_source, parameters_source):
        """For YML, the source it the file path"""
        with open(config_source) as config_source_file:
            config_raw = config_source_file.read()

            parameters = {}
            # Parameters from file
            if os.path.isfile(parameters_source):
                params = self.load_parameters(parameters_source)
                if params is not None:
                    parameters.update(params)

            # Overwrite parameters with the environment variables
            env_params = {}
            env_params.update(os.environ)
            for key, value in env_params.items():
                if is_string(value):
                    env_params[key] = "'" + value + "'"

            parameters.update(env_params)
            # Replace the parameters
            final_configuration = config_raw.format(**parameters)
            final_configuration = yaml.safe_load(final_configuration)
            return final_configuration if final_configuration is not None else {}



"""Base events used by the applauncher kernel. Bundles events can extend them"""
import blinker


class EventHierarchy(type):
    """Build the full namespace when creating an event"""
    def __new__(cls, name, bases, dct):
        signal_list = [dct["event_name"]]
        for base in bases:
            if hasattr(base, "_signals"):
                signal_list += getattr(base, "_signals")
        dct["_signals"] = signal_list
        return type.__new__(cls, name, bases, dct)


class Event(metaclass=EventHierarchy):
    """Generic event"""
    event_name = "event"


class KernelReadyEvent(Event):
    """Raised when the kernel is fully ready"""
    event_name = "kernel.kernel_ready"


class ConfigurationReadyEvent(Event):
    """Raised when the configuration is ready. The configuration is provided in the event"""
    event_name = "kernel.configuration_ready"

    def __init__(self, configuration):
        self.configuration = configuration


class InjectorReadyEvent(Event):
    """Raised when you can use the dependency injection container"""
    event_name = "kernel.injector_ready"


class KernelShutdownEvent(Event):
    """This event will be sent to your bundle to notify that it should stop everything"""
    event_name = "kernel.kernel_shutdown"


class EventManager:
    """To register and dispatch application events"""
    @staticmethod
    def add_listener(event, listener):
        """Listen for an event"""
        if isinstance(event, str):
            channel = blinker.signal(event)
        else:
            channel = blinker.signal(event.event_name)

        channel.connect(listener)

    @staticmethod
    def dispatch(event):
        """This event will be propagated to all listeners"""
        for channel in getattr(event, "_signals"):
            blinker.signal(channel).send(event)



"""Applaucher kernel and boot process"""
import sys
import signal
import importlib
import types
from rich.traceback import install
from rich.table import Table
from rich.console import Console
from pydantic import ValidationError
from dependency_injector import containers, providers
from .logging import configure_logger
from .service_runner import ProcessServiceRunner
from .event import KernelReadyEvent, KernelShutdownEvent, ConfigurationReadyEvent
from .event import EventManager
from .configuration import load_configuration


install()
configure_logger()


class Configuration(providers.Provider):
    """Configuration injector"""
    def __deepcopy__(self, memo):
        copied = memo.get(id(self))
        if copied is not None:
            return copied

        copied = self.__class__()
        self._copy_overridings(copied, memo)

        return copied

    def _provide(self, _args, _kwargs):
        return Kernel.config


class ServiceContainerMeta(type):
    """Service container use interface. Here all containers will be gathered"""
    def __getattr__(cls, key):
        if cls.container is None:
            raise Exception("Service container not configured yet!")
        if not hasattr(cls.container, key):
            raise Exception(f'Service container does not have any "{key}" service')
        return getattr(cls.container, key)


class ServiceContainer(metaclass=ServiceContainerMeta):
    """This service container will contain all bundle and application containers"""
    container = None


class Kernel:
    """The application initializer. It loads all the components and then let the control to the bundles"""
    inject_bindings = {}
    config = None

    container = containers.DynamicContainer()

    def load_configuration(self, configuration_file, parameters_file):
        """Parse the configuration and put it into the kernel and container"""
        self.console.log("Parsing configuration...")
        try:
            self.config = load_configuration(
                configuration_file_path=configuration_file,
                parameters_file_path=parameters_file,
                bundles=self.bundles
            )
            Kernel.config = self.config
            self.container.configuration = Configuration()
            self.console.log("Configuration [bold green]OK[/]")
        except ValidationError as ex:
            self.console.rule("[bold red]Configuration error")
            for error in ex.errors():
                loc = "[yellow] -> [/]".join([f"[cyan]{i}[/]" for i in error["loc"]])
                self.console.print(f"{loc}: {error['msg']} (type={error['type']})")
            sys.exit()

    def kernel_ready_event(self, _event):
        """Seeing this message is the proof that the events are working"""
        table = Table(show_header=False, style="bold green")
        table.add_row("Kernel ready")
        self.console.print(table)

    def kernel_shutdown_event(self, _event):
        """Seeing this message is the proof that the events are working"""
        table = Table(show_header=False, style="bold cyan")
        table.add_row("Kernel shutdown")
        self.console.print(table)

    def register_event_listeners(self):
        """Add all kernel and bundles listeners"""
        self.console.log("Registering event listeners")
        self.event_manager.add_listener(event=KernelReadyEvent, listener=self.kernel_ready_event)
        self.event_manager.add_listener(event=KernelShutdownEvent, listener=self.kernel_shutdown_event)
        for bundle in self.bundles:
            if hasattr(bundle, 'event_listeners'):
                for event_type, listener in bundle.event_listeners:
                    self.event_manager.add_listener(event=event_type, listener=listener)
                event_list = ', '.join(
                    [f'[bright_magenta]{event_type.__name__}[/]' for event_type, _ in bundle.event_listeners]
                )
                self.console.log(f"Registered events for [bold cyan]{bundle.__class__.__name__}[/]: {event_list}")
        self.console.log("Events [bold green]OK[/]")

    def build_dependency_container(self):
        """Build the kernel and bundles service containers"""
        self.console.log("Building dependency container")
        wire_modules = []
        for bundle in self.bundles:
            if hasattr(bundle, 'injection_bindings'):
                for class_type, provider in bundle.injection_bindings.items():
                    # A container provider is useful when a container has dependencies on other containers. Just provide
                    # the container class as an easy way to use the defaults
                    if isinstance(provider, types.FunctionType):
                        # We use a function instead of directly the Provider to ensure that the dependencies are using
                        # exactly the same container (and not other instance). Otherwise, the singletons will not be
                        # singletons (will be at least two instances, one per container
                        provider = provider(self.container)
                    else:
                        provider = providers.Container(provider)
                    setattr(self.container, class_type, provider)

            if hasattr(bundle, 'wire_modules'):
                wire_modules += bundle.wire_modules
        if wire_modules:
            self.console.log(f"[bold cyan]Wiring[/] modules: {wire_modules}")
            self.container.wire(modules=[importlib.import_module(module) for module in wire_modules])
        self.console.log("Dependency container [bold]built[/]")

    def register_services(self):
        """Register the bundles services but not run them yet, it will be done later"""
        self.console.log("Registering services")
        for bundle in self.bundles:
            if hasattr(bundle, 'services'):
                for service_name, service_function, args, kwargs in bundle.services:
                    self.console.log(f"Adding {service_name}")
                    self.service_runner.add_service(
                        name=service_name,
                        function=service_function,
                        args=args,
                        kwargs=kwargs
                    )

    def __init__(self,
                 environment,
                 bundles,
                 configuration_file="config/config.yml",
                 parameters_file="config/parameters.yml"):
        self.console = Console()
        self.console.log(f"Running environment [bold green]{environment}[/]")
        self.bundles = bundles
        self.environment = environment
        self.event_manager = EventManager()
        self.service_runner = ProcessServiceRunner()
        self.shutting_down = False

        # Signals
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)

        with self.console.status("[bold green]Booting kernel..."):
            self.load_configuration(configuration_file=configuration_file, parameters_file=parameters_file)
            self.register_event_listeners()
            self.event_manager.dispatch(ConfigurationReadyEvent(configuration=self.config))
            self.build_dependency_container()
            self.event_manager.dispatch(KernelReadyEvent())
            self.register_services()

        self.service_runner.run()

    def _signal_handler(self, _os_signal, _frame):
        self.shutdown()

    def shutdown(self):
        """Start the kernel shutdown process. It will stopp all services and the exit"""
        if not self.shutting_down:
            table = Table(show_header=False, style="bold red")
            table.add_row("Shutdown signal received, press ctrl + c to kill the process")
            self.console.print(table)
            self.shutting_down = True
            self.service_runner.shutdown()
            self.container.shutdown_resources()
        else:
            self.console.log("[bold red]Killing...[/]")
            self.service_runner.kill()

    def wait(self):
        """Wait until all services are done"""
        self.service_runner.wait()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.event_manager.dispatch(KernelShutdownEvent())


ServiceContainer.container = Kernel.container



"""Service runner, like multiprocessing but with more features"""
from multiprocessing import Process
import asyncio
import os
import logging


class ProcessServiceRunner:
    """Like a process pool but with initializer and friendly exit process"""
    def __init__(self):
        self.logger = logging.getLogger("service")
        self.running_services = []

    def add_service(self, name, function, args=None, kwargs=None):
        """Register a function that will run in the background"""
        if args is None:
            args = ()

        if kwargs is None:
            kwargs = {}

        self.running_services.append((name, Process(target=function, args=args, kwargs=kwargs)))

    def run(self):
        """Start all registered services"""
        for name, process in self.running_services:
            self.logger.info("Starting service %s", name)
            process.start()

    def wait(self):
        """Wait until all services have finished"""
        self.logger.info("Waiting for services to end")
        for _, i in self.running_services:
            i.join()

    def kill(self):
        """Send a stop signal to all services"""
        for name, process in self.running_services:
            self.logger.info("Killing service %s (%s)", name, process.pid)
            os.kill(process.pid, 9)
            process.join()

    async def _terminate_processes(self, processes, grace_time=10):
        """Trying to finish all services concurrently"""
        await asyncio.gather(*[
            self._terminate_process(*i, grace_time) for i in processes
        ])

    async def _terminate_process(self, name, process, grace_time=10):
        """Try to finish a service. First attempt is friendly, after the grace time it will be killed"""
        self.logger.info("Terminating service %s (%s)", name, process.pid)
        process.terminate()
        wait = True
        while wait:
            grace_time -= 1
            await asyncio.sleep(1)
            wait = process.is_alive() and grace_time > 0

        if process.is_alive():  # pragma: no cover  signals cannot be tested with pytest, manual test required
            self.logger.info("Killing service %s (%s)", name, process.pid)
            os.kill(process.pid, 9)
            process.join()

    def shutdown(self, grace_time=10):
        """Start the shutdown process"""
        self.logger.info("Shutting down services (grace time of %s seconds)", grace_time)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._terminate_processes(self.running_services, grace_time))



"""Fancy logging"""
import logging
from rich.logging import RichHandler


def configure_logger():
    """Default log format"""
    log_header = "%(asctime)s [bold cyan]%(levelname)s[/] [yellow]-[/] [royal_blue1]%(name)s[/] [yellow]-[/]"
    log_body = "%(message)s [yellow]([/][chartreuse4]%(filename)s[/]:%(lineno)d[yellow])[/]"
    log_format = f"{log_header} {log_body}"
    logging.basicConfig(
        level=logging.INFO, format=log_format, datefmt="[%X]", handlers=[
            RichHandler(show_time=False, show_level=False, markup=True, tracebacks_show_locals=True,
                        rich_tracebacks=True)]
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['applauncher', 'applauncher.cli', 'applauncher.cli.generate']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'blinker>=1.4,<2.0',
 'dependency-injector>=4.20.1,<5.0.0',
 'pydantic>=1.7.3,<2.0.0',
 'rich>=9.10.0,<10.0.0',
 'typer>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'applauncher',
    'version': '2.0.1',
    'description': 'Python application launcher',
    'long_description': None,
    'author': 'Alvaro Garcia',
    'author_email': 'maxpowel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
